<?php
//Global variables
$baseUrl;
$baseFolderName;
$flag = true;
$firstURL = 'http://localhost/repository.git/info/refs?service=git-upload-pack';
/**
 * This call reproduces the effects of Git clone operation using cURL requests
 * @author Snigdha Rao Parvatneni snigdha.parvatneni@gmail.com
 * @copyright 2013
 */

class Git
{
    /**
     * This function will make cURL call to execute first git clone call
     * @param null
     * @return null
     */
    function getGitRepository()
    {
        //variable declaration
        global $baseUrl;
        global $firstURL;
        //display message to the user
        echo 'Cloning started'."\n";
        //cURL call
        $content = get_data($firstURL);
        //extracting common url
        $pos = strpos($firstURL, ".git");
        $baseUrl = substr($firstURL, 0, $pos + 4);
        //calling createBaseFolder method to create the base folder
        Git::createBaseFolder();
        //calling getGitMasterFile method to get data from master file
        Git::getGitMasterFile($content);
    }

    /**
     * This function will get the name of base folder to clone the files
     * @param null
     * @return null
     */
    function createBaseFolder() 
    {
        //variable declaration
        global $baseUrl;
        global $baseFolderName;
        global $flag;
        $elements = array();
        //exploding the string
        $elements = explode("/", $baseUrl);
        //to get the name of the base folder
        for($i=0; $i < count($elements); $i++) {
            $pos = strpos($elements[$i], ".git");
            if(is_bool($pos) !== true) {
                $baseFolderName = substr($elements[$i], 0, $pos);
            }
        }
        //creating base folder
        if (!file_exists($baseFolderName)) {
            mkdir($baseFolderName, 0777, true);
        } else {
            $flag = false;
        }
    }

    /**
     * This function will get the content of master file
     * @param string $content it contains  the uncompressed contents of first
              git request
     * @return null
     */
    function getGitMasterFile($content)
    {
        //variable declaration
        global $baseUrl;
        //making url for next git clone call
        $extendedUrl = substr($content, 0, 40);
        $firstSplit = substr($extendedUrl, 0, 2);
        $secondSplit = substr($extendedUrl, 2, 38);
        $connect = $firstSplit."/".$secondSplit;
        $url = $baseUrl."/objects/".$connect;
        //making cURL call to the url
        $compContent = get_data($url);
        $uncompContent = gzuncompress($compContent);
        //calling getGitMasterTree to read the tree object
        $url = Git::getGitMasterTree($uncompContent);
    }

    /**
     * This function will make the url for the master tree object
     * @param string $content it contains the uncompressed contents of first
              git request
     * @return null
     */
     function getGitMasterTree($content)
     {
          //variable declaration
          global $baseUrl;
          //forming url
          $treeName = substr($content, 16, 41);
          $firstSplit = substr($treeName, 0, 2);
          $secondSplit = substr($treeName, 2, 38);
          $connect = $firstSplit."/".$secondSplit;
          $url = $baseUrl."/objects/".$connect;
          /*calling getGitMasterTreeStructure to read the structure inside the
            master tree object*/
          Git::getGitMasterTreeStructure($url);
    }

    /**
     * This function will get the structure of tree objects
     * @param string $url it contains the url to make the second request for
              git clone
     * @return null
     */
    function getGitMasterTreeStructure($url) 
    {
        //declarations
        global $baseFolderName;
        //making a cURL call
        $compContent = get_data($url);
        $uncompContent = gzuncompress($compContent);
        //call to another method
        Git::getObjects($uncompContent, $baseFolderName);
    }

    /**
     * This function will get the git objects
     * @param string $content it contains the uncompressed contents of third
              request for git clone
     * @param string $path it contains path of the git object requested
     * @return null
     */
    function getObjects($content, $path) 
    {
        //declarations
        global $baseUrl;
        //to read the contents of git clone
        $tempContent[1] = $content;
        for($i = 0; $i < strlen($content); $i++) {
            //getting positions of first apprearence of access codes
            $blobPos = strpos($tempContent[1], '100644');
            $treePos = strpos($tempContent[1], '40000');
            $extraBlobPos = strpos($tempContent[1], '100755');
            //calling method checkPosition
            $element = Git::checkPosition($blobPos, $treePos, $extraBlobPos ,
                                          $content);
            //resetting the values
            $blobPos = $element[0];
            $treePos = $element[1];
            $extraBlobPos = $element[2];
            if($blobPos < $treePos && $blobPos < $extraBlobPos) {
                $tempContent = Git::readBlobSha($tempContent[1], $blobPos,
                                                strlen($tempContent[1]), $path);
            }
            else if($treePos < $blobPos || $treePos < $extraBlobPos) {
                $tempContent = Git::readTreeSha($tempContent[1], $treePos,
                                                strlen($tempContent[1]), $path);
            }
            else if($extraBlobPos < $treePos && $extraBlobPos < $blobPos) {
                $tempContent = Git::readBlobSha($tempContent[1], $extraBlobPos,
                                                strlen($tempContent[1]), $path);
            }
            $i = strlen($tempContent[1]);
            $i = strlen($content) - $i;
        }
    }

    /**
     * This function check the position of access code for null values
     * @param integer $blobPos it is the position of firt occurence of 100644
     * @param integer $treePos it is the position of firt occurence of 40000
     * @param integer $extraBlobPos it is the position of first occurence of
              100755
     * @param string $content it contains the uncompressed contents of from the
              request for git clone
     * @return array $element its value represents first occurence of blobPos,
               treePos and extraBlobPos
     */
    function checkPosition($blobPos, $treePos, $extraBlobPos, $content) 
    {
        $element = array();
        if(is_bool($blobPos) === true) {
            $blobPos = strlen($content);
        }
        if(is_bool($treePos) === true) {
            $treePos = strlen($content);
        }
        if(is_bool($extraBlobPos) === true) {
            $extraBlobPos = strlen($content);
        }
        $element[0] = $blobPos;
        $element[1] = $treePos;
        $element[2] = $extraBlobPos;
        return $element;
    }

    /**
     * This function is used to read the content from the blob file
     * @param string $content it contains the uncompessed contents of from the
              request for git clone
     * @param integer $blobPos it is the position of firt occurence of 100644
     * @param integer $length it is the length of the compressed content
     * @param string $path it is the path of the blob object requested
     * @return array $returnContent it is the actual content written to the file
     */
    function readBlobSha($content, $blobPos, $length, $path)
    {
        $returnContent = array();
        $values = array();
        $tempContent = substr($content, $blobPos, $length);
        $accessCode = substr($tempContent, 0, 6);
        $values[0] = $accessCode;
        $tempContent = substr($tempContent, 7, $length);
        $tempPos = strpos($tempContent, "\x00");
        $fileName = substr($tempContent, 0, $tempPos);
        $values[1] = $fileName;
        $tempContent = substr($tempContent, $tempPos + 1, $length);
        $shaBinary = substr($tempContent, 0, 20);
        $shaHash = bin2hex($shaBinary);
        $values[2] = $shaHash;
        $tempContent = substr($tempContent, 20, $length);
        $returnContent[0] = $values;
        $returnContent[1] = $tempContent;
        $returnContent[2] = $path;
        Git::writeContent($shaHash, $path, $fileName);
        return $returnContent;
    }

    /**
     * This function is used to read the content from the tree object
     * @param string $content it contains the uncompressed contents of from the
              request for git clone
     * @param integer $treePos it is the position of firt occurence of 40000
     * @param integer $length it is the length of the compressed content
     * @param string $path it is the path of the tree object requested
     * @return array $returnContent it is the actual content of the tree object
     */
    function readTreeSha($content, $treePos, $length, $path)
    {
        $returnContent = array();
        $values = array();
        $tempContent = substr($content, $treePos, $length);
        $accessCode = substr($tempContent, 0, 5);
        $values[0] = $accessCode;
        $tempContent = substr($tempContent, 6, $length);
        $tempPos = strpos($tempContent, "\x00");
        $folderName = substr($tempContent, 0, $tempPos);
        $values[1] = $folderName;
        $tempContent = substr($tempContent, $tempPos + 1, $length);
        $shaBinary = substr($tempContent, 0, 20);
        $shaHash = bin2hex($shaBinary);
        $values[2] = $shaHash;
        $tempContent = substr($tempContent, 20, $length);
        $returnContent[0] = $values;
        $returnContent[1] = $tempContent;
        $returnContent[2] = $path;
        $path = $path.'/'.$folderName;
        Git::createFolder($path);
        Git::checkNestedStructure($shaHash, $path);
        return $returnContent;
    }

    /**
     * This function is used to check nested folder structure
     * @param string $shaHash it contains the SHA hash of the Git object
     * @param string $path it is the path of the tree object requested
     * @return null
     */
    function checkNestedStructure($shaHash, $path) 
    {
        $url = Git::urlMaker($shaHash);
        //making cURL call
        $compContent = get_data($url);
        $uncompContent = gzuncompress($compContent);
        Git::getObjects($uncompContent, $path);
    }

    /**
     * This function is used to make url from SHA hash
     * @param string $shaHash it contains the SHA hash of the Git object
     * @return string $url it is the actual url to access the git object
     */
    function urlMaker($shaHash) 
    {
        //variable declaration
        global $baseUrl;
        $firstSplit = substr($shaHash, 0, 2);
        $secondSplit = substr($shaHash, 2, 38);
        $connect = $firstSplit."/".$secondSplit;
        $url = $baseUrl."/objects/".$connect;
        return $url;
    }

    /**
     * This function is function is used to create folders
     * @param string $path it is the path of the tree object requested
     * @return null
     */
    function createFolder($path)
    {
        //variable declaration
        global $flag;
        if (!file_exists($path)) {
            //echo "The file $baseFolderName exists"."\n";
            mkdir($path, 0777, true);
        } else {
            $flag = false;
        }
    }

    /**
     * This function is function is used to write the contents to the file
     * @param string $shaHash it the SHA hash of the git object
     * @param string $path it is the path of the tree object requested
     * @param string $fileName it is the actual name of the file
     * @return null
     */
    function writeContent($shaHash, $path, $fileName)
    {
        $url = Git::urlMaker($shaHash);
        $compContent = get_data($url);
        $uncompContent = gzuncompress($compContent);
        $length = strlen ($uncompContent);
        $pos = strpos($uncompContent, "\x00");
        $fileContent = substr($uncompContent, $pos + 1, $length);
        $outFileHandle = fopen($path.'/'.$fileName, 'w')
                         or die("can't open file");
        fwrite($outFileHandle, $fileContent);
        fclose($outFileHandle);
    }

    /**
     * This function is function is used display message to the screen
     * @param null
     * @return null
     */
    function displayMessage()
    {
        global $flag;
        if($flag != false) {
            echo 'Cloning completed'."\n";
        } else {
            echo 'Folder already exists'."\n";
        }
    }
}
//end of class

/** 
* This function is an anynomyous function taken from internet to make the cURL
  request
* @param string $url it the actual url for making cURL call
* @return string $data it is the compressed data received from making cURL
          request.
*/
function get_data($url) {
    $ch = curl_init();
    $timeout = 5;
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, $timeout);
    $data = curl_exec($ch);
    curl_close($ch);
    return $data;
}

//creating Git object
$git = new Git();
$git->getGitRepository();
$git->displayMessage();
?>