package com.hmk.invertedindex;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeMap;


/**
 * COPYRIGHT (C) 2012 Mallika Perepa, Snigdha rao Parvatneni and Chinmayee Annachhatre. All Rights Reserved.
 * CS 267 -02  Homework #1
 * 
 * Inverted Index Printer
 * 
 * @author Mallika Perepa-008335301  Snigdha rao Parvatneni-008648978    Chinmayee Annachhatre-008150285
 * @version 1.0 Sep 11, 2012
 */
public class IndexPrinter {
	/**
	 * This is the main method of the IndexPrinter class. 
	 * @param args	String[]
	 */
	public static void main (String args[]) {
		if(isOkToProceed(args)){
            try{
                List<String> contents = readFile(args[0]);
                Map<String, Word> sortedMap = countWords(contents);
                printResult(sortedMap);   
            }catch(FileNotFoundException e){
                System.err.println("File not found. Please input a valid file");
            }
		}
	}
    
    /**
     * Checks if the user entered file is valid or not.
     *
     * @param args	String[]
     * @return boolean
     */
    private static boolean isOkToProceed(String args[]){
        boolean value = true;
        if(args.length == 0 || null == args[0]){
			System.err.println("Invalid filename");
			value = false;
		}
        
        return value;
    }
    
    /**
     * process each document and extracts words from it and creates a Word object and stores 
     * the word and document counts.
     *
     * @param contents List<String>
     * @return Map<String, Word>
     */
    private static Map<String, Word> countWords(List<String> contents){
        Map<String, Word> sortedMap = new TreeMap<String, IndexPrinter.Word>();
		int docCount = 0;
		for (String content : contents) {
			content = content.replaceAll("\n"," ");
			String[] values = content.split(" ");
			for (String value : values) {
				if(sortedMap.containsKey(value.trim())){
					Word word = sortedMap.get(value.trim());
					word.setTotalCount(word.getTotalCount() + 1);
					word.getDocCount()[docCount]++;
					sortedMap.put(value.trim(), word);
				}else{
					Word word =new IndexPrinter().new Word();
					word.setTotalCount(1);
					word.setWord(value.trim());
					word.getDocCount()[docCount] = 1;
					sortedMap.put(value.trim(), word);
				}
			}
			docCount++;
		}
        return sortedMap;
    }
    
    /**
     * Print the counts in the following format,
     * number of documents, number of total occurances, (document number, number of occurences in that particular document),...
     *
     * @param sortedMap Map<String, Word>
     */
    private static void printResult(Map<String, Word> sortedMap){
        Iterator<String> iter = sortedMap.keySet().iterator();
		while(iter.hasNext()){
			Word value = sortedMap.get(iter.next());
			int[] docCounts = value.getDocCount();
			int numberOfDocs = 0;
			StringBuilder frequenciesToBePrinted = new StringBuilder();
			int index = 0;
			for (int i : docCounts) {
				index++;
				if(i > 0){
					numberOfDocs++;
					frequenciesToBePrinted.append("(");
					frequenciesToBePrinted.append(index);
					frequenciesToBePrinted.append(",");
					frequenciesToBePrinted.append(i);
					frequenciesToBePrinted.append("),");
				}
			}
			frequenciesToBePrinted.deleteCharAt(frequenciesToBePrinted.length()  - 1);
			System.out.println(value.getWord() + "\n " + numberOfDocs + "," +value.getTotalCount() + "," + frequenciesToBePrinted);
		}
    }
	
	/**
	 * This method reads the file, extract the documents from it and store each document in a List object.
     
	 * @param fileName	String
	 * @return	List<String>
	 * @throws FileNotFoundException
	 */
	private static List<String> readFile(String fileName) throws FileNotFoundException{
		List<String> contents = new ArrayList<String>();
		Scanner scanner = new Scanner(new File(fileName));
		scanner.useDelimiter("\n\n");
		while(scanner.hasNext()){
			String text = scanner.next().trim();
			if(null != text && text.length() > 0){
                contents.add(text);	
			}
		}
		return contents;
	}
     
    /**
     * This is a inner class to hold the Word data.
     *
     */
	private class Word{
		private String word;
		private int totalCount = 0;
		private int[] docCount = new int[50000];
		public String getWord() {
			System.out.println("word from word class----"+word);
			return word;
		}
		public void setWord(String word) {
			this.word = word;
		}
		public int getTotalCount() {
			System.out.println("get total count from word----"+totalCount);
			return totalCount;
		}
		public void setTotalCount(int totalCount) {
			System.out.println("set total count from word----"+totalCount);
			this.totalCount = totalCount;
		}
		public int[] getDocCount() {
			System.out.println("get doc count----"+docCount);
			return docCount;
		}
	}
}
