package com.hmk.invertedindex;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Arrays;
import java.util.TreeMap;
import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;

/**
  COPYRIGHT (C) 2013 Snigdha rao Parvatneni. All Rights Reserved. CS 297 Deliverable 2
  Naive based classifier to recognize between Java and Python
  @author Snigdha rao Parvatneni-008648978
  @version 1.0 March 19, 2012
 */
public class TrigramIndex {
	
	//Global Declarations
	static TreeMap<String, Word> sortedJavaMap = new TreeMap<String, TrigramIndex.Word>();
	static TreeMap<String, Word> sortedPythonMap = new TreeMap<String, TrigramIndex.Word>();
	static TreeMap<String, Word> sortedRandomJavaMap = new TreeMap<String, TrigramIndex.Word>();
	static TreeMap<String, Word> sortedRandomPythonMap = new TreeMap<String, TrigramIndex.Word>();
	static ArrayList<String> queryTrigrams = new ArrayList<String>();	
	static BigDecimal javaUnknown;
	static BigDecimal pythonUnknown;

	/**
	  This is the main method of the TrigramIndex class.
	  This method calls other methods.
	  @param args	String[]
	 */
	public static void main (String args[]) 
	{
		if(isOKToProceed(args)){
			try{
				//method call to read txt files
	            List<String> javaContents = readFile(args[0]);
	            List<String> pythonContents = readFile(args[1]);
	            List<String> javaRandomContents = readFile(args[2]);
	            List<String> pythonRandomContents = readFile(args[3]);
	            String query = args[4];          
	           //method call to split query into trigrams 
	           queryTrigrams = splitQuery(query);	   	           
	           //method call to save trigrams in TreeMap
	           sortedJavaMap = (TreeMap<String, Word>) countWords(javaContents);	  
	           sortedPythonMap = (TreeMap<String, Word>) countWords(pythonContents);
	           sortedRandomJavaMap = (TreeMap<String, Word>) countWords(javaRandomContents);
	           sortedRandomPythonMap = (TreeMap<String, Word>) countWords(pythonRandomContents);         
	           //method call to calculate probabilities 
	           initialProbability(javaContents,sortedJavaMap);
	           initialProbability(pythonContents,sortedPythonMap);
	           BigDecimal javaLambda = unknownProbability(sortedJavaMap,sortedRandomJavaMap);
	           javaUnknown = javaLambda;	           
	           BigDecimal pythonLambda = unknownProbability(sortedPythonMap,sortedRandomPythonMap);
	           pythonUnknown = pythonLambda;
	   		   BigDecimal one = new BigDecimal("1.00");
	           javaLambda = one.subtract(javaLambda);
	           pythonLambda = one.subtract(pythonLambda);        
	           smoothnedProbability(sortedJavaMap, javaLambda);
	           smoothnedProbability(sortedPythonMap, pythonLambda);
	           searchQuery(sortedJavaMap, sortedPythonMap, queryTrigrams);
	           
			}catch(FileNotFoundException e){
	            System.err.println("File not found. Please input a valid file");
	        }
		}
	}
	
    /**
    This method checks if the trigrams in the query appears in Java or Python.
    @param sortedJavaMap	Map<String, Word>
    @param sortedPythonMap	Map<String, Word>
    @param queryTrigrams	ArrayList<String>
    */
	private static void searchQuery(Map<String, Word> sortedJavaMap, Map<String, Word> sortedPythonMap, 
			ArrayList<String> queryTrigrams)
	{
		//Declarations
		ArrayList<String> matchedJavaTrigrams = new ArrayList<String>();
		ArrayList<String> matchedPythonTrigrams = new ArrayList<String>();
		ArrayList<String> unknownJavaTrigrams = new ArrayList<String>();
		ArrayList<String> unknownPythonTrigrams = new ArrayList<String>();
		//size of query
		int size = queryTrigrams.size();
		BigDecimal javaResult;
		BigDecimal pythonResult;
		
		for(int i =0;i<size;i++){
			if(sortedJavaMap.containsKey(queryTrigrams.get(i))){
				matchedJavaTrigrams.add(queryTrigrams.get(i));
			}
			if(sortedPythonMap.containsKey(queryTrigrams.get(i))){
				matchedPythonTrigrams.add(queryTrigrams.get(i));
			}
			if(!sortedJavaMap.containsKey(queryTrigrams.get(i))){
				unknownJavaTrigrams.add(queryTrigrams.get(i));
			}
			if(!sortedPythonMap.containsKey(queryTrigrams.get(i))){
				unknownPythonTrigrams.add(queryTrigrams.get(i));
			}
		}	
		
		//method call to calculate the probabilities
		javaResult = calculateJavaProbability(matchedJavaTrigrams, unknownJavaTrigrams);
		pythonResult = calculatePythonProbability(matchedPythonTrigrams, unknownPythonTrigrams);
		
		System.out.println("Probability (query is from Java)------->"+ javaResult);
		System.out.println("Probability (query is from Python)----->"+ pythonResult);
		classifyResult(javaResult,pythonResult);		
	}
	
    /**
    This method prints the message in console that query belongs to Java or Python.
    This method prints the result based on comparison of results obtained for Java and Python.
    @param sortedJavaMap	Map<String, Word>
    @param sortedPythonMap	Map<String, Word>
    @param queryTrigrams	ArrayList<String>
    */
	private static void classifyResult(BigDecimal javaResult, BigDecimal pythonResult)
	{		
		//initialize the result
		int result = -2;
		result = javaResult.compareTo(pythonResult);
		
		//if else block to print the message
		if(result == 1){
			System.out.println("Result----->Query entered belongs to Java programming language");
		}
		else if(result == -1){
			System.out.println("Result----->Query entered belongs to Python programming language");
		}
		else if (result == 0){
			System.out.println("Result----->Query entered belongs to Both programming languages");
		}
		else{
			System.out.println("Result----->Query entered neither belongs to Java programming language nor belongs to Python programming language");
		}
	}
	
	/**
    This method calculate the probability that trigrams present in the query are from Java.
    @param matchedJavaTrigrams	ArrayList<String>
    @param matchedPythonTrigrams	ArrayList<String>
    @param unknownTrigrams	ArrayList<String>
    @return	BigDecimal
    */
	private static BigDecimal calculateJavaProbability(ArrayList<String> matchedJavaTrigrams, ArrayList<String> unknownJavaTrigrams)
	{
		int javaSize = matchedJavaTrigrams.size();
		int unknownSize = unknownJavaTrigrams.size();
		ArrayList<BigDecimal> javaProb = new ArrayList<BigDecimal>();	

		//fetching probability of trigrams present in Java
		for(int i=0; i<javaSize; i++){
			javaProb.add(sortedJavaMap.get(matchedJavaTrigrams.get(i)).getSmoothnedProb().get(matchedJavaTrigrams.get(i)));
		}
		
		//fetching probability of unknown trigrams
		for(int i=0; i<unknownSize; i++){
			javaProb.add(javaUnknown);
		}
		return multiplyProbability(javaProb, "Java");
	}
	
	/**
    This method calculate the probability that trigrams present in the query are from Python.
    @param matchedJavaTrigrams	ArrayList<String>
    @param matchedPythonTrigrams	ArrayList<String>
    @param unknownTrigrams	ArrayList<String>
    @return	BigDecimal
    */
	private static BigDecimal calculatePythonProbability(ArrayList<String> matchedPythonTrigrams, ArrayList<String> unknownPythonTrigrams)
	{
		int pythonSize = matchedPythonTrigrams.size();
		int unknownSize = unknownPythonTrigrams.size();
		ArrayList<BigDecimal> pythonProb = new ArrayList<BigDecimal>();
		BigDecimal one = new BigDecimal("1.00");
		
		//fetching probability of trigrams present in Python
		for(int i=0; i<pythonSize; i++){
			pythonProb.add(sortedPythonMap.get(matchedPythonTrigrams.get(i)).getSmoothnedProb().get(matchedPythonTrigrams.get(i)));
		}
		
		//fetching probability of unknown trigrams
		for(int i=0; i<unknownSize; i++){
			pythonProb.add(pythonUnknown);
		}
		
		return multiplyProbability(pythonProb, "Python");
	}
	
	/**
    This method multiply all the probabilities that required to classify the query as Java or Python.
    @param probabilityList	ArrayList<String>
    @param language	String
    @return	BigDecimal
    */
	private static BigDecimal multiplyProbability(ArrayList<BigDecimal> probabilityList, String language)
	{	
		BigDecimal ALPHA = new BigDecimal("1.00");
		BigDecimal intermediate = new BigDecimal("1.00");;
		BigDecimal result = new BigDecimal("1.00");
		BigDecimal hypProb = new BigDecimal("1.00");
		MathContext mc = new MathContext(100);
		
		int size = probabilityList.size();
		
		//loop to multiply probabilities of trigrams as decided previously
		for(int j=0;j<size;j++){
			intermediate = probabilityList.get(j);
			result = intermediate.multiply(result, mc);
		}
		
		//calling method to calculate probabilities of hypothesis Java and Python separately
		hypProb = hypothesisProbability(language);
		result = ((result.multiply(ALPHA, mc)).multiply(hypProb, mc));
		
		return result;
	}
	
	/**
    This method calculates the probabilities of hypothesis Java and Python separately
    @param language	String
    @return	BigDecimal
    */
	private static BigDecimal hypothesisProbability(String language)
	{	
		BigDecimal countJava = new BigDecimal("177000000.00");
		BigDecimal countPython= new BigDecimal("42000000.00");
		BigDecimal resultProb = new BigDecimal("1.00");
		
		BigDecimal countTotal = countJava.add(countPython); 		
		BigDecimal javaProb = countJava.divide(countTotal, 100, RoundingMode.HALF_UP);
		BigDecimal pythonProb = countPython.divide(countTotal, 100, RoundingMode.HALF_UP);
		
		//check for Java and Python
		if(language.equalsIgnoreCase("Java")){
			resultProb = javaProb;
		}
		if(language.equalsIgnoreCase("Python")){
			resultProb = pythonProb;
		}
		return resultProb;
	}
	
	/**
    This method calculates the initial probabilities of trigrams in training set
    @param contents	List<String>
    @param sortedMap Map<String, Word>
    */
	private static void initialProbability(List<String> contents, Map<String, Word> sortedMap)
	{
		int documentCount = contents.size();
		BigDecimal docCount = new BigDecimal(documentCount);
		
		//initial probabilities are given by count(no of docs which has trigram)/total no of docs in training set
		for(Map.Entry<String,Word> entry : sortedMap.entrySet()) {
			  String key = entry.getKey();
			  int frequency = count(entry.getValue().getPos().get(key));
			  BigDecimal freqCount = new BigDecimal(frequency);
			  BigDecimal iniProb = freqCount.divide(docCount, 100, RoundingMode.HALF_UP);
			  Map<String, BigDecimal> initialProb = new HashMap<String, BigDecimal>();
			  initialProb.put(key, iniProb);
			  sortedMap.get(key).setInitialProb(initialProb);
			}	
	}
	
	/**
    This method calculates the probabilities of trigrams not present in the training set
    @param sortedMap	Map<String, Word>
    @param sortedRandomMap  Map<String, Word>
    @return BigDecimal
    */
	private static BigDecimal unknownProbability(Map<String, Word> sortedMap, Map<String, Word> sortedRandomMap)
	{	
		BigDecimal unknownCount = new BigDecimal("0.00");
		BigDecimal one = new BigDecimal("1.00");
		BigDecimal trigramCount = new BigDecimal(sortedRandomMap.size());
		BigDecimal unknownProb = new BigDecimal("0.00");
		
		int size = sortedRandomMap.keySet().toArray().length;
		for(int i=0; i<size; i++){
			String key = sortedRandomMap.keySet().toArray()[i].toString();
			if(!sortedMap.containsKey(key)){
				unknownCount = unknownCount.add(one);
			}
		}	
		//precision used for division is 100
		unknownProb = unknownCount.divide(trigramCount,100, RoundingMode.HALF_UP);
		return unknownProb;
	}
	
	/**
    This method smoothen the probabilities of trigrams present in the training set
    @param sortedMap	Map<String, Word>
    @param smootheningFactor  BigDecimal
    */
	private static void smoothnedProbability(Map<String, Word> sortedMap, BigDecimal smootheningFactor)
	{
		BigDecimal iniProb;
		BigDecimal smoothProb;
		MathContext mc = new MathContext(100);
		
		//multiplying initial probabilities with smoothening factor
		for(Map.Entry<String,Word> entry : sortedMap.entrySet()) {
			String key = entry.getKey();
			iniProb = sortedMap.get(key).getInitialProb().get(key);
			smoothProb = iniProb.multiply(smootheningFactor,mc);
			Map<String, BigDecimal> smoothnedProb = new HashMap<String, BigDecimal>();
			smoothnedProb.put(key, smoothProb);
			sortedMap.get(key).setSmoothnedProb(smoothnedProb);
		}
	}
	
	/**
    This method splits the query into trigrams
    @param query  String
    @return ArrayList<String>
    */
	private static ArrayList<String> splitQuery(String query)
	{	
		ArrayList<String> qrTrigrams = new ArrayList<String>();		
		
		//splliting the trigrams
		if(query.length()>3){
			for(int i=0; i<query.length()-2; i++){
				String trigrams = query.substring(i,i+3); 
				if(!qrTrigrams.contains(trigrams)){
					qrTrigrams.add(trigrams);
				}
			}
		}else{
			String trigrams = query;
			if(!qrTrigrams.contains(trigrams)){
				qrTrigrams.add(trigrams);
			}
		}
		return qrTrigrams;
	}
	
	/**
    Checks if the user entered file names and query is valid or not.
    @param args	String[]
    @return boolean
  	*/
	private static boolean isOKToProceed(String args[])
	{
        boolean value = true;
        if(args.length == 0 || null == args[0] || null == args[1] 
                 || null == args[2] || null == args[3])
        {
			System.err.println("Invalid filenames");
			value = false;
		}
        
        if(null == args[4] || args[4].trim().length()==0){
			System.err.println("Invalid Query");
			value = false;
		}        
        return value;
    }
	
    /**
    This method reads the file, extract the documents from it and store each document in a List object.
    @param fileName	String
    @return	List<String>
    @throws FileNotFoundException
    */
	private static List<String> readFile(String fileName) throws FileNotFoundException
	{
		List<String> contents = new ArrayList<String>();
		Scanner scanner = new Scanner(new File(fileName));
		scanner.useDelimiter("\n\n");
		while(scanner.hasNext()){
			String text = scanner.next().trim();
			if(null != text && text.length() > 0){
                contents.add(text);	
			}
		}
		return contents;
	}
	
	private static int count(int[] array) {
        int count = 0;
        for(int i : array){
            if(i >= 0){
	            count++;
            }
        }
        return count;
    }
	
   /**
    This method process each document and extracts trigrams.
    @param contents List<String>
    @return Map<String, Word>
   */
   private static Map<String, Word> countWords(List<String> contents)
   {
	  //declarations
	  TreeMap<String, Word> sortedMap = new TreeMap<String, TrigramIndex.Word>();
	  int docNumber = -1;
	  for(String document : contents) {
		  docNumber = docNumber+1;
		  //splitting document per line
		  String[] details = document.split("\n");
		  for (String detail : details) {
			  if(detail.length()>3){
				  for(int i=0; i<detail.length()-2 ;i++){
					//splitting into trigrams  
					String trigrams = detail.substring(i, i+3); 
					  if(sortedMap.containsKey(trigrams)){	  
						  addExistingTrigrams(sortedMap, trigrams, docNumber);
					  }
					  else{
						  addNewTrigrams(sortedMap, trigrams, docNumber);
					  }
				  }
			  }
			  else{
				  if(sortedMap.containsKey(detail)){
					  addExistingWords(sortedMap, detail, docNumber);
				  }
				  else{
					  addNewWords(sortedMap, detail, docNumber);
				  }
			  }
		  }
	  }
	  return sortedMap;	  
   }
   
   /**
   This method add the document number of existing word of size 
   smaller or equal to 3 to inner class Word
   @param sortedMap TreeMap<String, Word>
   @param detail String
   @param docNumber int
   @return Map<String, Word>
  */
   private static Map<String, Word> addExistingWords(TreeMap<String, Word> sortedMap, 
		   String detail, int docNumber)
   { 
	  //adds docNumber to the already existing word
	  Word shortWord = sortedMap.get(detail);
	  int [] location = sortedMap.get(detail).getPos().get(detail);
	  for(int k=0; k<count(location); k++){
		  if(location[k] != docNumber && k == count(location)-1){
			  location[k+1] = docNumber; 
			  shortWord.getPos().put(detail, location);
			  sortedMap.put(detail, shortWord);
		  }
	  } 
	   return sortedMap;
   }
   
   /**
   This method add new words and document number of words of size 
   smaller or equal to 3 to inner class Word
   @param sortedMap TreeMap<String, Word>
   @param detail String
   @param docNumber int
   @return Map<String, Word>
  */
   private static Map<String, Word> addNewWords(TreeMap<String, Word> sortedMap,String detail, int docNumber)
   { 
	  //adding new word and docNumber of that word
	  Word shorterWord =new TrigramIndex().new Word();
	  int [] loc = new int [5000];
	  Arrays.fill(loc, -1);
	  shorterWord.setWord(detail);
	  loc[0] = docNumber;
	  shorterWord.getPos().put(detail, loc);
	  sortedMap.put(detail, shorterWord);
	  
	  return sortedMap;
   }
   
   /**
   This method add the document number of existing trigrams to inner class Word
   @param sortedMap TreeMap<String, Word>
   @param trigrams String
   @param docNumber int
   @return Map<String, Word>
  */
   private static Map<String, Word> addExistingTrigrams(TreeMap<String, Word> sortedMap,String trigrams, int docNumber)
   { 
	  //adding docNumbers for existing trigrams 
	  Word words = sortedMap.get(trigrams);
	  int [] position = sortedMap.get(trigrams).getPos().get(trigrams);
	  for(int k=0; k<count(position); k++){
		  if(position[k] != docNumber && k == count(position)-1){
			  position[k+1] = docNumber; 
			  words.getPos().put(trigrams, position);
			  sortedMap.put(trigrams, words);
		  }
	  }
	   return sortedMap;
   }
   
   /**
   This method add new trigrams and document number of trigrams to inner class Word
   @param sortedMap TreeMap<String, Word>
   @param trigrams String
   @param docNumber int
   @return Map<String, Word>
  */
   private static Map<String, Word> addNewTrigrams(TreeMap<String, Word> sortedMap, String trigrams, int docNumber)
   {	   	  
      Word word =new TrigramIndex().new Word();
	  int [] pos = new int [5000];
	  Arrays.fill(pos, -1);
	  word.setWord(trigrams);
	  pos[0] = docNumber;
	  word.getPos().put(trigrams, pos);
	  sortedMap.put(trigrams, word);  
	  
	  return sortedMap;
   }
	
    /**
	This is a inner class to hold the Word data.
	*/
	private class Word{
		private String word;
		private Map<String, int[]> pos = new HashMap<String, int[]>();
		private Map<String, BigDecimal> initialProb = new HashMap<String, BigDecimal>();
		private Map<String, BigDecimal> smoothnedProb = new HashMap<String, BigDecimal>();
		public String getWord() {
			return word;
		}
		public void setWord(String word) {
			this.word = word;
		}
		public Map<String, int[]> getPos() {
			return pos;
		}
		public void setPos(Map<String, int[]> pos) {
			this.pos = pos;
		}
		public Map<String, BigDecimal> getInitialProb() {
			return initialProb;
		}
		public void setInitialProb(Map<String, BigDecimal> initialProb) {
			this.initialProb = initialProb;
		}
		public Map<String, BigDecimal> getSmoothnedProb() {
			return smoothnedProb;
		}
		public void setSmoothnedProb(Map<String, BigDecimal> smoothnedProb) {
			this.smoothnedProb = smoothnedProb;
		}
	}
}

