package com.hmk.invertedindex;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Arrays;
import java.util.TreeMap;
import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;

public class TrigramIndex {
	//Global Declarations
	static TreeMap<String, Word> sortedJavaMap = new TreeMap<String, TrigramIndex.Word>();
	static TreeMap<String, Word> sortedPythonMap = new TreeMap<String, TrigramIndex.Word>();
	static TreeMap<String, Word> sortedRandomJavaMap = new TreeMap<String, TrigramIndex.Word>();
	static TreeMap<String, Word> sortedRandomPythonMap = new TreeMap<String, TrigramIndex.Word>();
	static ArrayList<String> queryTrigrams = new ArrayList<String>();	
	static BigDecimal javaUnknown;
	static BigDecimal pythonUnknown;

	
	//Starting of main method
	public static void main (String args[]) {
		if(isOKToProceed(args)){
			try{
	            List<String> javaContents = readFile(args[0]);
	            List<String> pythonContents = readFile(args[1]);
	            List<String> javaRandomContents = readFile(args[2]);
	            List<String> pythonRandomContents = readFile(args[3]);
	            String query = args[4];
	            
	            queryTrigrams = splitQuery(query);	            
	            sortedJavaMap = (TreeMap<String, Word>) countWords(javaContents);	  
	            sortedPythonMap = (TreeMap<String, Word>) countWords(pythonContents);
	            sortedRandomJavaMap = (TreeMap<String, Word>) countWords(javaRandomContents);
	            sortedRandomPythonMap = (TreeMap<String, Word>) countWords(pythonRandomContents); 
	            
	           initialProbability(javaContents,sortedJavaMap);
	           initialProbability(pythonContents,sortedPythonMap);
	           BigDecimal javaLambda = unknownProbability(sortedJavaMap,sortedRandomJavaMap);
	           javaUnknown = javaLambda;	           
	           BigDecimal pythonLambda = unknownProbability(sortedPythonMap,sortedRandomPythonMap);
	           pythonUnknown = pythonLambda;
	   		   BigDecimal one = new BigDecimal("1.00");
	           javaLambda = one.subtract(javaLambda);
	           pythonLambda = one.subtract(pythonLambda);
	           System.out.println("#################javaLambda###########"+javaUnknown);
	           System.out.println("#################pythonLambda###########"+pythonUnknown);

	           smoothnedProbability(sortedJavaMap, javaLambda);
	           smoothnedProbability(sortedPythonMap, pythonLambda);
	           searchQuery(sortedJavaMap, sortedPythonMap, queryTrigrams);
	           
			}catch(FileNotFoundException e){
	            System.err.println("File not found. Please input a valid file");
	        }
		}
	}
	
	private static void searchQuery(Map<String, Word> sortedJavaMap, Map<String, Word> sortedPythonMap, ArrayList<String> queryTrigrams){
		ArrayList<String> matchedJavaTrigrams = new ArrayList<String>();
		ArrayList<String> matchedPythonTrigrams = new ArrayList<String>();
		ArrayList<String> unknownTrigrams = new ArrayList<String>();
		int size = queryTrigrams.size();
		BigDecimal javaResult;
		BigDecimal pythonResult;
		String status;
		
		for(int i =0;i<size;i++){
			if(sortedJavaMap.containsKey(queryTrigrams.get(i))){
				matchedJavaTrigrams.add(queryTrigrams.get(i));
			}
			else if(sortedPythonMap.containsKey(queryTrigrams.get(i))){
				matchedPythonTrigrams.add(queryTrigrams.get(i));
			}
			else{
				unknownTrigrams.add(queryTrigrams.get(i));
			}
		}	
		javaResult = calculateJavaProbability(matchedJavaTrigrams, matchedPythonTrigrams, unknownTrigrams);
		pythonResult = calculatePythonProbability(matchedJavaTrigrams, matchedPythonTrigrams, unknownTrigrams);
		
		System.out.println("javaResult***************"+ javaResult);
		System.out.println("pythonResult***************"+ pythonResult);

		status = classifierResult(javaResult,pythonResult);
		
		if(status.equalsIgnoreCase("Java")){
			System.out.println("Querry entered belongs to Java programming language");
		}
		else if(status.equalsIgnoreCase("Python")){
			System.out.println("Querry entered belongs to Python programming language");
		}
		else if(status.equalsIgnoreCase("Both")){
			System.out.println("Querry entered belongs to Both programming language");
		}
		else if(status.equalsIgnoreCase("None")){
			System.out.println("Querry entered neither belongs to Java programming language nor belongs to Python programming language");
		}
	}
	
	private static String classifierResult(BigDecimal javaResult, BigDecimal pythonResult){
		String status;
		int result = -2;
		result = javaResult.compareTo(pythonResult);
		System.out.println("result classifier------->"+result);
		
		if(result == 1){
			status = "Java";
		}
		else if(result == -1){
			status = "Python";
		}
		else if (result == 0){
			status = "Both";	
		}
		else{
			status = "None";
		}
		return status;
	}
	
	private static BigDecimal calculateJavaProbability(ArrayList<String> matchedJavaTrigrams, ArrayList<String> matchedPythonTrigrams, ArrayList<String> unknownTrigrams){
		int javaSize = matchedJavaTrigrams.size();
		int pythonSize = matchedPythonTrigrams.size();
		int unknownSize = unknownTrigrams.size();
		ArrayList<BigDecimal> javaProb = new ArrayList<BigDecimal>();	
		BigDecimal one = new BigDecimal("1.00");

		//System.out.println("*****java*****"+matchedJavaTrigrams.size());
		//System.out.println("*****python*****"+matchedPythonTrigrams.size());
		//System.out.println("*****unknown*****"+unknownTrigrams.size());
		
		for(int i=0; i<javaSize; i++){
			System.out.println(matchedJavaTrigrams.get(i) +sortedJavaMap.get(matchedJavaTrigrams.get(i)).getSmoothnedProb().get(matchedJavaTrigrams.get(i)));
			javaProb.add(sortedJavaMap.get(matchedJavaTrigrams.get(i)).getSmoothnedProb().get(matchedJavaTrigrams.get(i)));
		}
		
		for(int i=0; i<pythonSize; i++){
			System.out.println(matchedPythonTrigrams.get(i) +(one.subtract(sortedPythonMap.get(matchedPythonTrigrams.get(i)).getSmoothnedProb().get(matchedPythonTrigrams.get(i)))));
			javaProb.add((one.subtract(sortedPythonMap.get(matchedPythonTrigrams.get(i)).getSmoothnedProb().get(matchedPythonTrigrams.get(i)))));
		}
		for(int i=0; i<unknownSize; i++){
			System.out.println(unknownTrigrams.get(i) + javaUnknown);
			javaProb.add(javaUnknown);
		}
		return multiplyProbability(javaProb, "Java");
	}
	
	private static BigDecimal calculatePythonProbability(ArrayList<String> matchedJavaTrigrams, ArrayList<String> matchedPythonTrigrams, ArrayList<String> unknownTrigrams){
		
		int javaSize = matchedJavaTrigrams.size();
		int pythonSize = matchedPythonTrigrams.size();
		int unknownSize = unknownTrigrams.size();
		ArrayList<BigDecimal> pythonProb = new ArrayList<BigDecimal>();
		BigDecimal one = new BigDecimal("1.00");
		
		System.out.println("########java########"+matchedJavaTrigrams.size());
		System.out.println("########python########"+matchedPythonTrigrams.size());
		System.out.println("########unknown########"+unknownTrigrams.size());
		
		for(int i=0; i<pythonSize; i++){
			System.out.println("In python \n");
			System.out.println(matchedPythonTrigrams.get(i) +sortedPythonMap.get(matchedPythonTrigrams.get(i)).getSmoothnedProb().get(matchedPythonTrigrams.get(i)));
			pythonProb.add(sortedPythonMap.get(matchedPythonTrigrams.get(i)).getSmoothnedProb().get(matchedPythonTrigrams.get(i)));
		}
		
		for(int i=0; i<javaSize; i++){
			System.out.println("In java \n");
			System.out.println(matchedJavaTrigrams.get(i) +(one.subtract(sortedJavaMap.get(matchedJavaTrigrams.get(i)).getSmoothnedProb().get(matchedJavaTrigrams.get(i)))));
			pythonProb.add((one.subtract(sortedJavaMap.get(matchedJavaTrigrams.get(i)).getSmoothnedProb().get(matchedJavaTrigrams.get(i)))));
		}
		for(int i=0; i<unknownSize; i++){
			System.out.println("In neither \n");
			System.out.println(unknownTrigrams.get(i) + pythonUnknown);
			pythonProb.add(javaUnknown);
		}
		return multiplyProbability(pythonProb, "Python");
	}
	
	private static BigDecimal multiplyProbability(ArrayList<BigDecimal> probabilityList, String language){
		
		BigDecimal ALPHA = new BigDecimal("1.00");
		BigDecimal intermediate = new BigDecimal("1.00");;
		BigDecimal result = new BigDecimal("1.00");
		BigDecimal hypProb = new BigDecimal("1.00");
		MathContext mc = new MathContext(100);
		
		int size = probabilityList.size();
		//System.out.println("size--------->"+size);
		
		for(int j=0;j<size;j++){
			intermediate = probabilityList.get(j);
			result = intermediate.multiply(result, mc);
			//System.out.println("loop result------->"+result+" language "+language);
		}
		//System.out.println("initial result------->"+result+" language "+language);
		hypProb = hypothesisProbability(language);
		//System.out.println("hypothesisProbability------->"+hypProb);
		
		result = ((result.multiply(ALPHA, mc)).multiply(hypProb, mc));
		//System.out.println("final result------->"+result+" language "+language);
		
		return result;
	}
	
	private static BigDecimal hypothesisProbability(String language){	
		
		BigDecimal countJava = new BigDecimal("177000000.00");
		BigDecimal countPython= new BigDecimal("42000000.00");
		BigDecimal resultProb = new BigDecimal("1.00");
		
		BigDecimal countTotal = countJava.add(countPython); 
		//System.out.println("countTotal-----*******--->"+countTotal);
		
		BigDecimal javaProb = countJava.divide(countTotal, 20, RoundingMode.HALF_UP);
		BigDecimal pythonProb = countPython.divide(countTotal, 20, RoundingMode.HALF_UP);
		
		//System.out.println("javaProb---*****-->"+javaProb);
		//System.out.println("pythonProb----*****->"+pythonProb);
		
		if(language.equalsIgnoreCase("Java")){
			resultProb = javaProb;
		}
		if(language.equalsIgnoreCase("Python")){
			resultProb = pythonProb;
		}
		return resultProb;
	}
	
	private static void initialProbability(List<String> contents, Map<String, Word> sortedMap){
		int documentCount = contents.size();
		BigDecimal docCount = new BigDecimal(documentCount);
		//System.out.println("docment count--------->"+documentCount);
		for(Map.Entry<String,Word> entry : sortedMap.entrySet()) {
			  String key = entry.getKey();
			  int frequency = count(entry.getValue().getPos().get(key));
			  BigDecimal freqCount = new BigDecimal(frequency);
			  BigDecimal iniProb = freqCount.divide(docCount, 20, RoundingMode.HALF_UP);
			  //System.out.println("key----->"+key+ "frequency------->"+frequency+" initial frequency----->"+iniProb);
			  Map<String, BigDecimal> initialProb = new HashMap<String, BigDecimal>();
			  initialProb.put(key, iniProb);
			  sortedMap.get(key).setInitialProb(initialProb);
			}	
	}
	
	private static BigDecimal unknownProbability(Map<String, Word> sortedMap, Map<String, Word> sortedRandomMap){
		
		BigDecimal unknownCount = new BigDecimal("0.00");
		BigDecimal one = new BigDecimal("1.00");
		BigDecimal trigramCount = new BigDecimal(sortedRandomMap.size());
		BigDecimal unknownProb = new BigDecimal("0.00");
		
		int size = sortedRandomMap.keySet().toArray().length;
		for(int i=0; i<size; i++){
			String key = sortedRandomMap.keySet().toArray()[i].toString();
			if(!sortedMap.containsKey(key)){
				unknownCount = unknownCount.add(one);
				//System.out.println("key-------->"+key);
			}
		}		
		unknownProb = unknownCount.divide(trigramCount,20, RoundingMode.HALF_UP);
		//System.out.println("#################unknownProb###########"+unknownProb);
		return unknownProb;
	}
	
	private static void smoothnedProbability(Map<String, Word> sortedMap, BigDecimal smootheningFactor){
		BigDecimal iniProb;
		BigDecimal smoothProb;
		for(Map.Entry<String,Word> entry : sortedMap.entrySet()) {
			String key = entry.getKey();
			iniProb = sortedMap.get(key).getInitialProb().get(key);
			smoothProb = iniProb.multiply(smootheningFactor);
			Map<String, BigDecimal> smoothnedProb = new HashMap<String, BigDecimal>();
			smoothnedProb.put(key, smoothProb);
			sortedMap.get(key).setSmoothnedProb(smoothnedProb);
		}
	}
	
	private static ArrayList<String> splitQuery(String query){
		ArrayList<String> qrTrigrams = new ArrayList<String>();		
		if(query.length()>3){
			for(int i=0; i<query.length()-3; i++){
				String trigrams = query.substring(i,i+3); 
				if(!qrTrigrams.contains(trigrams)){
					qrTrigrams.add(trigrams);
					//System.out.println("query trigrams in if--------->"+trigrams);
				}
			}
		}else{
			String trigrams = query;
			if(!qrTrigrams.contains(trigrams)){
				qrTrigrams.add(trigrams);
				//System.out.println("query trigrams in else--------->"+trigrams);
			}
		}
		return qrTrigrams;
	}
	/**
    Checks if the user entered file name, query name and query algorithm is valid or not.
    @param args	String[]
    @return boolean
  */
	//working
	private static boolean isOKToProceed(String args[]){
        boolean value = true;
        if(args.length == 0 || null == args[0] || null == args[1]){
			System.err.println("Invalid filenames");
			value = false;
		}
        
        if(null == args[1] || args[1].trim().length()==0){
			System.err.println("Invalid Query");
			value = false;
		}        
        return value;
    }
	/**
  	This method reads the file, extract the documents from it and store each document in a List object.
  	@param fileName	String
  	@return	List<String>
 	@throws FileNotFoundException
   */
	//working
	private static List<String> readFile(String fileName) throws FileNotFoundException{
		List<String> contents = new ArrayList<String>();
		Scanner scanner = new Scanner(new File(fileName));
		scanner.useDelimiter("\n\n");
		while(scanner.hasNext()){
			String text = scanner.next().trim();
			if(null != text && text.length() > 0){
                contents.add(text);	
			}
		}
		return contents;
	}
	
	private static int count(int[] array) {
        int count = 0;
        for(int i : array){
            if(i >= 0){
	            count++;
            }
        }
        return count;
    }
	  /*
    process each document and extracts words from it and creates a Word object and stores 
    the word and document counts.
    @param contents List<String>
    @return Map<String, Word>
   */
  private static Map<String, Word> countWords(List<String> contents){
	  //declarations
	  TreeMap<String, Word> sortedMap = new TreeMap<String, TrigramIndex.Word>();
	  
	  //int documentCount = contents.size();
	  //System.out.println("content.size--------->"+contents.size());
	  int docNumber = -1;
	  
	  //document ---> each java program
	  for(String document : contents) {
		  docNumber = docNumber+1;
		  //System.out.println("******************Document Number---->"+docNumber+"***********************\n");
		  String[] details = document.split("\n");
		  //detail------>each line per program
		  for (String detail : details) {
			  //when length of a line is greater than 3
			  if(detail.length()>3){
				  //for entire line
				  for(int i=0; i<detail.length()-3 ;i++){
					String trigrams = detail.substring(i, i+3); 
					  if(sortedMap.containsKey(trigrams)){
						  Word words = sortedMap.get(trigrams);
						  int [] position = sortedMap.get(trigrams).getPos().get(trigrams);
						  for(int k=0; k<count(position); k++){
							  if(position[k] != docNumber && k == count(position)-1){
								  position[k+1] = docNumber; 
								  words.getPos().put(trigrams, position);
								  sortedMap.put(trigrams, words);
							  }
						  } 
					  }
					  else{
						  //System.out.println("trigrams in else "+trigrams);
						  Word word =new TrigramIndex().new Word();
						  int [] pos = new int [5000];
						  Arrays.fill(pos, -1);
						  word.setWord(trigrams);
						  pos[0] = docNumber;
						  //System.out.println("trigrams---->"+trigrams +" docnumber------->"+pos[0]);
						  word.getPos().put(trigrams, pos);
						  sortedMap.put(trigrams, word);
					  }
				  }
			  }
			  else{
				  //System.out.println("detail------>"+detail);
				  if(sortedMap.containsKey(detail)){
					  Word shortWord = sortedMap.get(detail);
					  int [] location = sortedMap.get(detail).getPos().get(detail);
					  for(int k=0; k<count(location); k++){
						  if(location[k] != docNumber && k == count(location)-1){
							  location[k+1] = docNumber; 
							  shortWord.getPos().put(detail, location);
							  sortedMap.put(detail, shortWord);
						  }
					  } 
				  }
				  else{
					  //System.out.println("trigrams in else "+trigrams);
					  Word shorterWord =new TrigramIndex().new Word();
					  int [] loc = new int [5000];
					  Arrays.fill(loc, -1);
					  shorterWord.setWord(detail);
					  loc[0] = docNumber;
					  //System.out.println("trigrams---->"+trigrams +" docnumber------->"+pos[0]);
					  shorterWord.getPos().put(detail, loc);
					  sortedMap.put(detail, shorterWord);
				  }
			  }
		  }
	  }
	  return sortedMap;	  
  }
	/**
	This is a inner class to hold the Word data.

	*/
	private class Word{
		private String word;
		private Map<String, int[]> pos = new HashMap<String, int[]>();
		private Map<String, BigDecimal> initialProb = new HashMap<String, BigDecimal>();
		private Map<String, BigDecimal> smoothnedProb = new HashMap<String, BigDecimal>();
		public String getWord() {
			return word;
		}
		public void setWord(String word) {
			this.word = word;
		}
		public Map<String, int[]> getPos() {
			return pos;
		}
		public void setPos(Map<String, int[]> pos) {
			this.pos = pos;
		}
		public Map<String, BigDecimal> getInitialProb() {
			return initialProb;
		}
		public void setInitialProb(Map<String, BigDecimal> initialProb) {
			this.initialProb = initialProb;
		}
		public Map<String, BigDecimal> getSmoothnedProb() {
			return smoothnedProb;
		}
		public void setSmoothnedProb(Map<String, BigDecimal> smoothnedProb) {
			this.smoothnedProb = smoothnedProb;
		}
	}
}

